////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any
//     purpose is hereby granted without fee, provided that the above copyright
//     notice appear in all copies and that both that copyright notice and this
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software
//     for any purpose. It is provided "as is" without express or implied
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_PRIVILIGEDACCESS_H_INC_
#define REFLIB_PRIVILIGEDACCESS_H_INC_


#include <boost/type_traits/add_reference.hpp>

#include "../qualifiedtype.h"

#include "../issameclass.h"

namespace refl
{
    // We borrow from Smalltalk the terms ClassVariable and InstanceVariable 
    // because they are clearer then static member variable and non-static 
    // member variable
    struct InstanceVariableTag{};

    struct ClassVariableTag{};

    // traversion might be static members only 
    // (in which case no record instance is needed)
    // non-static members only or both.
    struct TraverseInstanceAndClassVariables{};

    struct KindIsMatch{};

    struct KindIsNoMatch{};

    template<class AttibuteKind, class MemberKinds> struct AttributeKindMatches
    {
        enum {value = boost::is_same<AttibuteKind,MemberKinds>::value};  
        typedef typename boost::ct_if<(value != 0),
                            KindIsMatch,
                            KindIsNoMatch >::type
            type;
    };
    
    template <class AttributeKind> struct AttributeKindMatches<AttributeKind,
                                              TraverseInstanceAndClassVariables>
    {
        typedef KindIsMatch type;
    };

    namespace detail
    {
        // check for instance Herb Sutters article on "befriending templates"
        // http://www.cuj.com/documents/s=8244/cujcexp2101sutter/sutter.htm
        // why we choose this form (Note that we do want to support VC++ 6.0)
        struct PrivilegedAccessImpl
        {
            // if no kinds are specified:
            // return references to static and non-static membervariables
            template<class ReflectedClass, class Tag, class  ValueType> static
            typename boost::add_reference<ValueType>::type
            getValueHandle(ReflectedClass& recordData, Tag)
            {
                return recordData.template getByTag<Tag>();
            }
            // return references to static OR non-static membervariables
            template<class ReflectedClass, class Tag, class  ValueType,
                    class MemberKind> static
            typename boost::add_reference<ValueType>::type
            getValueHandle(ReflectedClass& recordData, Tag tag,
                        MemberKind kind)
            {
                return recordData.getByTag(tag,kind);
            }
            // workaround for not so perfect partial ordering of function
            // templates with gcc 3.2
            template<class ReflectedClass, class Tag, class  ValueType,
                    class MemberKind> static
            typename boost::add_reference<ValueType>::type
            getValueHandle(ReflectedClass& recordData, Tag tag,
                        MemberKind kind, IsNotSameClassTag)
            {
                return recordData.getByTag(tag,kind);
            }
            // return references to static membervariables
            template<class ReflectedClass, class Tag, class  ValueType> static
            typename boost::add_reference<ValueType>::type
            getValueHandle(Tag tag)
            {
                return getStaticValueHandle<ReflectedClass,Tag,
                                    ValueType>(tag, IsConst<ValueType>::type());
            }
            // for static member functions the trick to overload on the 
            // constness of *this cannot work - hence the function that returns
            // a modifyable reference to a static member must be named 
            // differently (or overloaded otherwise).
            template<class ReflectedClass, class Tag, class  ValueType> static
            typename boost::add_reference<ValueType>::type
            getStaticValueHandle(Tag tag,TypeIsConstTag)
            {
                return ReflectedClass::getByTag(tag,ClassVariableTag());
            }
            template<class ReflectedClass, class Tag, class  ValueType> static
            typename boost::add_reference<ValueType>::type
            getStaticValueHandle(Tag tag,TypeIsNotConstTag)
            {
                return ReflectedClass::getRefByTag(tag,ClassVariableTag());
            }
            // if mixed kinds are specified:
            // return references to static and non-static membervariables
            template<class ReflectedClass, class Tag, class  ValueType> static
            typename boost::add_reference<ValueType>::type
            getValueHandle(ReflectedClass& recordData, Tag tag,
                        TraverseInstanceAndClassVariables, IsSameClassTag)
            {
                return recordData.template getByTag<Tag>();
            }
        };
        template<class ValueType, class MemberKinds> struct PrivilegedAccess
        {
            typedef typename boost::add_reference<ValueType>::type RefType;
            template<class ReflectedClass, class Tag> static
            RefType getValueHandle(ReflectedClass& recordData, Tag tag)
            {
                // unfortunately the trick does not work with gcc 3.2.
                typedef typename IsSameClass<MemberKinds,
                                TraverseInstanceAndClassVariables>::type
                        PartialOrderingBugWorkaround;
                return PrivilegedAccessImpl::template
                       getValueHandle<ReflectedClass,Tag,ValueType>(recordData,
                              tag,MemberKinds(),PartialOrderingBugWorkaround());
            }
            typedef typename boost::add_reference<ValueType>::type RefType;
            template<class ReflectedClass, class Tag> static
            RefType getValueHandle(Tag tag)
            {
                return PrivilegedAccessImpl::template
                           getValueHandle<ReflectedClass,Tag,ValueType>(tag);
            }
        };
    }
    template<class ReflectedClass, class MemberKinds, int Pos> struct Intruder
    {
        typedef typename ReflectedClass::template 
            GetNameTag<Pos>::type AttributeTag;
        
        typedef typename ReflectedClass::template 
            GetValueType<Pos>::type AttributeValueType;
        
        typedef typename ReflectedClass::template 
            GetMemberKind<Pos>::type AttributeKind;

        typedef typename GetQualifiedMemberType<ReflectedClass,
                    AttributeValueType>::type QualifiedValueType;
        typedef detail::PrivilegedAccess<QualifiedValueType,
                                        MemberKinds> MyIntruder;
        typedef typename MyIntruder::RefType RefType;
        typedef typename AttributeKindMatches<AttributeKind,MemberKinds>::type
            KindMatches;
        static RefType getValueHandle(ReflectedClass& recordData)
        {
            return MyIntruder::getValueHandle(recordData,AttributeTag());
        }
    };
    template<class ReflectedClass, int Pos> struct Intruder<ReflectedClass,
                                                ClassVariableTag,Pos>
    {
        typedef typename ReflectedClass::template 
            GetNameTag<Pos>::type AttributeTag;
        
        typedef typename ReflectedClass::template 
            GetValueType<Pos>::type AttributeValueType;
        
        typedef typename GetQualifiedMemberType<ReflectedClass,
                    AttributeValueType>::type QualifiedValueType;

        typedef typename ReflectedClass::template
            GetMemberKind<Pos>::type AttributeKind;

        typedef typename AttributeKindMatches<AttributeKind,
                                            ClassVariableTag>::type KindMatches;

        typedef detail::PrivilegedAccess<QualifiedValueType,
                                        ClassVariableTag> MyIntruder;
        typedef typename MyIntruder::RefType RefType;
        static RefType getValueHandle()
        {
            return MyIntruder
                ::template getValueHandle<ReflectedClass,AttributeTag>(
                                                                AttributeTag());
        }
        // this overload allows client code that usually accesses instance
        // variables to acccess class variables in the same way
        static RefType getValueHandle(ReflectedClass&)
        {
            return MyIntruder
                ::template getValueHandle<ReflectedClass,AttributeTag>(
                                                                AttributeTag());
        }
    };
    // having the attribute kind as a filter the need to generate no ops
    // when processing an attribute with non matching kind arises.
    // derived classes should define what happen when the kind matches:
    // template<class Function> static void
    // processMatchingAttribute(ReflectedClass& recordData,
    //                      Function& function)
    template<class ReflectedClass, class MemberKinds, int Pos, 
            class HoldsImplForMatchingKind> 
    struct MatchingAttributeHandler: Intruder<ReflectedClass,MemberKinds,Pos>
    {
        typedef Intruder<ReflectedClass,MemberKinds,Pos> Inherited;
        typedef typename Inherited::KindMatches KindMatches;
        template<class Function> static void
        processSingleAttribute(ReflectedClass& recordData,
                                Function& function)
        {  
            processSingleAttribute(recordData,function,KindMatches());
        } 
        // binary overload
        template<class Function> static void
        processSingleAttribute(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord,
                                Function& function)
        {  
            processSingleAttribute(lhsRecord, rhsRecord,function,KindMatches());
        } 
        template<class Function> static void 
        processSingleAttribute(ReflectedClass&,Function&,KindIsNoMatch)
        {  
            /* no op. 
            Happens when instance variables should be traversed and 
            the current attribute is a class variable (static member variable)
            */
        }   
        // binary overload
        template<class Function> static void 
        processSingleAttribute(ReflectedClass&,ReflectedClass&,Function&,KindIsNoMatch)
        {  
            /* no op. 
            Happens when instance variables should be traversed and 
            the current attribute is a class variable (static member variable)
            */
        }   
        template<class Function> static void 
        processSingleAttribute(ReflectedClass& recordData,Function& function,
                            KindIsMatch)
        {  
            HoldsImplForMatchingKind::processMatchingAttribute(recordData,
                                                              function);
        } 
        // binary overload
        template<class Function> static void 
        processSingleAttribute(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord, 
                            Function& function, KindIsMatch)
        {  
            HoldsImplForMatchingKind::processMatchingAttribute(lhsRecord,
                                                              rhsRecord, function);
        }   
    };
    // This version supports access for static members only when the caller
    // has no appropriate this pointer at hand.
    template<class ReflectedClass, int Pos, class HoldsImplForMatchingKind> 
    struct MatchingAttributeHandler<ReflectedClass,ClassVariableTag,Pos,
                        HoldsImplForMatchingKind>: 
                                Intruder<ReflectedClass,ClassVariableTag,Pos>
    {
        typedef Intruder<ReflectedClass,ClassVariableTag,Pos> Inherited;
        typedef typename Inherited::KindMatches KindMatches;
        template<class Function> static void 
        processSingleAttribute(Function& function)
        {  
            processSingleAttribute(function,KindMatches());
        } 
        template<class Function> static void 
        processSingleAttribute(Function&,KindIsNoMatch)
        {  
            /* no op. 
            Happens here when the current attribute is an instance variable 
            */
        }   
        template<class Function> static void 
        processSingleAttribute(Function& function,KindIsMatch)
        {  
            HoldsImplForMatchingKind::processMatchingAttribute(function);
        }
        // Note how radically this specialization avoids using an instance
        // of ReflectedClass
    };
}


#endif //
