////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software 
//     for any purpose. It is provided "as is" without express or implied 
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_SORTINGRECORDSBYATTRIBUTEVALUES_H_INC_
#define REFLIB_SORTINGRECORDSBYATTRIBUTEVALUES_H_INC_

#include <stdexcept>
#include <functional>
#include "../lexicographicalsort.h"
#include "../definereflection.h"
#include "../depthfirstforeach.h"

template<class Tag> struct SortRecordBy
{
    template <class RecordT> bool operator()(const RecordT& lhs, 
                                            const RecordT& rhs)const
    {
        return lhs.template getValueByTag<Tag>() < rhs.template getValueByTag<Tag>();
    }
};

template<class Tag> SortRecordBy<Tag> makeSortByTag(Tag)
{
    return SortRecordBy<Tag>();
}

template<class AttributeTag1, class AttributeTag2> 
LexicographicalSort<SortRecordBy<AttributeTag1>,SortRecordBy<AttributeTag2> >
makeLexicographicalRecordSort(AttributeTag1, AttributeTag2)
{
    return makeLexicographicalSort(SortRecordBy<AttributeTag1>(),
                                SortRecordBy<AttributeTag2>());

}

template<class AttributeTag1, class AttributeTag2, class AttributeTag3> 
    LexicographicalSort<SortRecordBy<AttributeTag1>,
        LexicographicalSort<SortRecordBy<AttributeTag2>, 
                            SortRecordBy<AttributeTag3> > >
makeLexicographicalRecordSort(AttributeTag1, AttributeTag2, AttributeTag3)
{
    return makeLexicographicalSort(SortRecordBy<AttributeTag1>(),
                                makeLexicographicalRecordSort(AttributeTag2(),
                                                            AttributeTag3()));
}

template<class AttributeTag1, class AttributeTag2, class AttributeTag3, 
        class AttributeTag4> 
    LexicographicalSort<SortRecordBy<AttributeTag1>,
        LexicographicalSort<SortRecordBy<AttributeTag2>,
            LexicographicalSort<SortRecordBy<AttributeTag3>, 
                            SortRecordBy<AttributeTag4> > > > 
makeLexicographicalRecordSort(AttributeTag1, AttributeTag2, 
                              AttributeTag3, AttributeTag4)
{
    return makeLexicographicalSort(SortRecordBy<AttributeTag1>(),
                        makeLexicographicalRecordSort(AttributeTag2(),
                                                    AttributeTag3(),
                                                    AttributeTag4()));
}

template<class AttributeTag1, class AttributeTag2, class AttributeTag3, 
        class AttributeTag4, class AttributeTag5> 
    LexicographicalSort<SortRecordBy<AttributeTag1>,
        LexicographicalSort<SortRecordBy<AttributeTag2>,
            LexicographicalSort<SortRecordBy<AttributeTag3>,
                LexicographicalSort<SortRecordBy<AttributeTag4>, 
                                    SortRecordBy<AttributeTag5 > > > > >
makeLexicographicalRecordSort(AttributeTag1, AttributeTag2, AttributeTag3, 
                              AttributeTag4, AttributeTag5)
{
    return makeLexicographicalSort(SortRecordBy<AttributeTag1>(),
                                makeLexicographicalRecordSort(AttributeTag2(),
                                                            AttributeTag3(),
                                                            AttributeTag4(),
                                                            AttributeTag5()));
}

namespace refl
{
    template<template<class> class ValueCmp = std::less> class CompareRecord
    {
    public:
        CompareRecord():resultKnown_(false),everyThingTested_(false),result_(false){};
        // was useful for printing base classes - and may be useful for
        // other binary traversions - only here it cannot do something
        // meaningful.
        template<class ReflectedClass> void 
        operator()(const ReflectedClass&, const ReflectedClass&)  
        {/* no op*/}

        // here ValueType is neither compiletime- nor runtime iterable.
        // Hence it can be compared "as is" or the user has to define an 
        // operator<(const ValueType& lhs, const ValueType& rhs)
        template<class Scope, class NameTag, class ValueType> 
        void operator()(const ValueType& lhsValue,const ValueType& rhsValue,
                        NameTag,const Scope* = NULL, const Scope* = NULL)
        {
            typedef ValueCmp<ValueType> Comparison;
            // if it was not performance critical then I would have used a setfunction for result_
            // that checks first whether the result is known.
            // Since this library already may overstress the inlining abilities of a compiler 
            // I skipped that.
            if(!resultKnown_)       
            {                        
                Comparison cmp;     
                result_ = cmp(lhsValue, rhsValue);
                resultKnown_ = result_ ||              // lhs precedes rhs 
                               cmp(rhsValue,lhsValue); // rhs precedes lhs 
            }
        }
        // this requires partial ordering of function templates.
        template<class Scope, class NameTag, class ValueType> 
        void operator()(const ValueType&,const ValueType&,NameTag,
                        LeaveAttributeTag,const Scope* = NULL, const Scope* = NULL)
        {
            enum {lastAttribute = (Scope::template AttributePos<NameTag>::value == Scope::NumAttributes - 1)};
            // in a nested structure of reflected classes this may be true when
            // in the outermost class still attributes are to check.
            // but this is no problem as long as no attempt is made to treat the comparison 
            // result at that stage as a comparison result for the outermost class.
            everyThingTested_ = lastAttribute;
        };        

        template<class Scope, class NameTag, class ValueType, class GraphEvent> 
        void operator()(const ValueType&,const ValueType&,NameTag,
                        GraphEvent,const Scope* = NULL, const Scope* = NULL)
        {/* no op*/}

        // lhs collection ran out of bullets
        template<class Scope, class NameTag, class ValueIter> 
        void operator()(PastEndTag,ValueIter /* rhsPos */,ValueIter /* rhsEnd */,
                        NameTag,const Scope* = NULL, const Scope* = NULL)
        {
            // lexicographical sort: "a" < "aa"
            if(!resultKnown_)
            {
                result_ = true;
                resultKnown_ = true;
            }
        }
        // rhs collection ran out of bullets
        template<class Scope, class NameTag, class ValueIter> 
        void operator()(ValueIter/* lhsPos */,ValueIter /* lhsEnd */, PastEndTag,
                        NameTag,const Scope* = NULL, const Scope* = NULL)
        {
            // lexicographical sort: "a" < "aa"
            if(!resultKnown_)
            {
                result_ = false;
                resultKnown_ = true;
            }
        }
        // lhs is a null pointer and rhs is not - hence it could be dereferenced
        template<class Scope, class NameTag, class ValueType> 
        void operator()(PastEndTag,const ValueType& /*rhsValue*/,
                        NameTag,const Scope* = NULL, const Scope* = NULL)
        {
            if(!resultKnown_)
            {
                // lexicographical sort: "a" < "aa"
                result_ = true;
                resultKnown_ = true;
            }
        }
        // rhs is a null pointer and lhs is not - hence it could be dereferenced
        template<class Scope, class NameTag, class ValueType> 
        void operator()(const ValueType& /*lhsValue*/,PastEndTag,
                        NameTag,const Scope* = NULL, const Scope* = NULL)
        {
            if(!resultKnown_)
            {
                // lexicographical sort: "a" < "aa"
                result_ = false;
                resultKnown_ = true;
            }
        }
        bool result()
        {
            if(!resultKnown_ && !everyThingTested_)
                throw std::logic_error("trying to access an undefined comparison result");
            return result_;
        }
        bool done()const
        {
            return resultKnown_;
        }
    private:
        bool everyThingTested_;
        bool result_;
        bool resultKnown_;
    };

    template<class Record> inline bool recordIsLess(const Record& lhs, 
                                                const Record& rhs)
    {
        CompareRecord<> comparison;
        depthFirstForeachIncAllBaseClasses(lhs, rhs, comparison, InstanceVariableTag()); 
        return comparison.result();
    };
}

template<class Record> struct RecordIsLess
{
    inline bool operator()(const Record& lhs, const Record& rhs)
    {
        return refl::recordIsLess(lhs,rhs);
    };
};

#endif // 

