////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software 
//     for any purpose. It is provided "as is" without express or implied 
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_USERCONFIG_H_INC_
#define REFLIB_USERCONFIG_H_INC_

#include <boost/utility.hpp>

// use this form if you prefer member variables with trailing underscore
#define REFLIB_MEMVARS_WITH_TRAILING_

// use this macro to customize the name of the setfunctions that contain
// the tagname as substring.
// Note that the templated set functions are not configurable because
// these may be used by generative algorithms (depthfirstforeach, ...).
// Catering for variable function names in generative algorithms 
// would add complexity where already is more than enough of that.
#define REFLIB_SET_FUNCTION_NAME(TagName) set##TagName

#define REFLIB_GET_FUNCTION_NAME(TagName) get##TagName
// the stl-style would be: (relying on overloading)
// #define REFLIB_SET_FUNCTION_NAME(TagName) TagName
// #define REFLIB_GET_FUNCTION_NAME(TagName) TagName


// use this form if you prefer member variables with leading m_
// #define REFLIB_MEMVARS_WITH_LEADING_m_
// you cannot use both forms
namespace refl
{
    struct SingleThreadingModelTag{};

    struct MultiThreadingModelTag{};

    struct EmptyMutex:boost::noncopyable {};

    // each lock shall be constructible with its corresponding mutex
    struct EmptyLock:boost::noncopyable 
    {
        EmptyLock(EmptyMutex&){}
    };
}


namespace userconfig
{
    // This can be useful for instance to apply additional checks on the new 
    // value 
    template<class ReflectedClass, class AttributeTag, class ValueType> 
    void preSet(ReflectedClass& rec, AttributeTag tag, ValueType val)
    { /* whatever needs to be done */}

    // the same for static member variables
    template<class ReflectedClass, class AttributeTag, class ValueType> 
    void preSet(AttributeTag tag, ValueType val)
    { /* whatever needs to be done */}

    // This was useful because the records did store the count of a container in
    // a separate variable setting the count-member therefore had to resize the
    // associated collection. (So this function can be used to ensure that post
    // conditions hold.)
    template<class ReflectedClass, class AttributeTag> void
    postSet(ReflectedClass& rec, AttributeTag)
    { /* whatever needs to be done */}

    // the same for static member variables
    template<class ReflectedClass, class AttributeTag> void
    postSet(AttributeTag)
    { /* whatever needs to be done */}

    
    typedef refl::SingleThreadingModelTag ThreadingModel;
    //typedef refl::MultiThreadingModelTag ThreadingModel;

    template<class ReflectedClass, class ThreadingModelType> struct MutexType;

    template<class ReflectedClass, class ThreadingModelType> struct LockType;
    
    // the settings for single threading might not need user adjustment
    template<class ReflectedClass> struct MutexType<ReflectedClass,
                                                refl::SingleThreadingModelTag>
    {
        typedef refl::EmptyMutex type;
    };

    template<class ReflectedClass> struct LockType<ReflectedClass,
                                                refl::SingleThreadingModelTag>
    {
        typedef refl::EmptyLock type;
    };

    // the settings for multi threading DO NEED USER ADJUSTMENT
    template<class ReflectedClass> struct MutexType<ReflectedClass,
                                                   refl::MultiThreadingModelTag>
    {
        typedef refl::EmptyMutex type;
    };

    template<class ReflectedClass> struct LockType<ReflectedClass,
                                                refl::MultiThreadingModelTag>
    {
        typedef refl::EmptyLock type;
    };

    // This one can be used to have one mutex per class (these mutexes will be 
    // used for setting static member variables.
    template<class ReflectedClass, class ThreadingModelType> class MutexInstance
    {
    public:
        typedef typename MutexType<ReflectedClass,ThreadingModelType>::type 
            Mutex;
        static Mutex& get();
    private:
        static Mutex mutex_;
    };
    template <class ReflectedClass, class ThreadingModelType> 
    typename MutexInstance<ReflectedClass,ThreadingModelType>::Mutex&
    MutexInstance<ReflectedClass,ThreadingModelType>::get()
    {
        return mutex_;
    }

    template <class ReflectedClass, class ThreadingModelType>
    typename MutexInstance<ReflectedClass,ThreadingModelType>::Mutex 
    MutexInstance<ReflectedClass,ThreadingModelType>::mutex_;
}

#endif //
