////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software 
//     for any purpose. It is provided "as is" without express or implied 
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_PRINTRECORDDESC_H_INC_
#define REFLIB_PRINTRECORDDESC_H_INC_

#ifdef _MSC_VER
#pragma warning(disable: 4267)  
#endif 

#include <iosfwd>

#include <sstream>

#include <string>

#include <vector>

#include <typeinfo>

#include <cassert>

#include "../issameclass.h"

#include "../depthfirstforeach.h"

#include "../depthfirstforeach.h"

#include "passtype.h"

namespace refl
{   
    class ObjectPath
    {
    public:
        void increaseDimension();
        void increaseDimension(const std::string&);
        void decreaseDimension();
        void prepareNext();
        void prepareNext(const std::string&);
        void reset();
        void nextInCollection();
    private:
        typedef std::vector<int>::size_type size_type;
        // one and only one of the items in that pair will be used
        typedef std::pair<size_type,const std::string*> NodeDesc;
        typedef std::vector<NodeDesc> Positions;
        Positions positions_;
        friend std::ostream& operator<<(std::ostream& out,
                                        const ObjectPath& index);
    };
    
    inline void ObjectPath::increaseDimension()
    {
        positions_.push_back(NodeDesc(0,NULL));
    }

    inline void ObjectPath::increaseDimension(const std::string& attributeName)
    {
        positions_.push_back(NodeDesc(0,&attributeName));
    }

    inline void ObjectPath::decreaseDimension()
    {
        assert(!positions_.empty());
        positions_.pop_back();
    }

    
    inline void ObjectPath::nextInCollection()
    {
        if(!positions_.back().second)
            positions_.back().first++; 
    }
    inline void ObjectPath::prepareNext()
    {
        assert(!positions_.empty());
        nextInCollection(); 
    }
    inline void ObjectPath::prepareNext(const std::string& attributeName)
    {
        assert(!positions_.empty());
        positions_.back().second = &attributeName;  
    }
    inline void ObjectPath::reset()
    {
        positions_.clear();  
    }

    inline std::ostream& operator<<(std::ostream& out,const ObjectPath& index)
    {
        typedef ObjectPath::Positions::const_iterator PosIter;
        if(index.positions_.empty())
            return out;
        
        PosIter posPos = index.positions_.begin();
        PosIter posEnd = index.positions_.end();
        if(!posPos->second)
            out << "[" << posPos->first << "]";
        else
            out << *(posPos->second);

        for(++posPos; posPos != posEnd; ++posPos)
            if(!posPos->second)
                out << "[" << posPos->first << "]";
            else
                out << "." << *(posPos->second);
        return out;
    }
    class PrintRecord
    {
    public:
        PrintRecord(std::ostream& out):stream_(out){}

        // overloads to traverse static members only having no instance of 
        // ReflectedClass
        // very bad hack for missing partial ordering of function templates
        template<class ReflectedClass> void print(PassType<ReflectedClass>, 
                                                ClassVariableTag)  
        {        
            // without an instance of a reflected class locking is done according
            // to the constness of ReflectedClass.
            // attribute functions that do not modify static member variables
            // should pass const types to the generative traversions 
            // (depthFirstForeach, allForeach, ...)
            typedef typename boost::add_const<ReflectedClass>::type ConstReflectedClass;
            path_.reset();
            path_.increaseDimension("");
            depthFirstForeachIncAllBaseClasses<ConstReflectedClass>(*this); 
            path_.decreaseDimension();
            stream_ << std::endl;
        }       
        // useful for printing base classes
        template<class ReflectedClass> void 
        operator()(PassType<ReflectedClass>, ClassVariableTag)  
        {        
            stream_ << ReflectedClass::getClassName() << ":\t";
        }


        // printing attributes when you have an instance of reflected class
        template<class ReflectedClass> void 
        print(const ReflectedClass& PrintThis)  
        {        
            path_.reset();
            path_.increaseDimension("");
            depthFirstForeachIncAllBaseClasses(PrintThis,*this,
                                        TraverseInstanceAndClassVariables()); 
            path_.decreaseDimension();
            stream_ << std::endl;
        }       
        // useful for printing base classes
        template<class ReflectedClass> void 
        operator()(const ReflectedClass& PrintThis)  
        {        
            stream_ << ReflectedClass::getClassName() << ":\t";
        }
        // here ValueType is neither compiletime- nor runtime iterable.
        // Hence it can be printed "as is" or the user has to define an 
        // operator<<(std::ostream&,const ValueType&)
        template<class WrappedScope, class NameTag, class ValueType> 
        void operator()(const ValueType& value,NameTag,WrappedScope)
        {
            stream_ << path_ << ": " << value << ";\t";
        }

        template<class WrappedScope, class NameTag, class ValueType> 
        void operator()(const ValueType&,NameTag, EnterReflectedClassTag, 
                        WrappedScope)
        {
            path_.increaseDimension("");    // quick hack - pushs an attribute 
                                            // dimension on the object path 
        }                                   // stack

        template<class WrappedScope, class NameTag, class ValueType> 
        void operator()(const ValueType&,NameTag , LeaveReflectedClassTag, 
                        WrappedScope)
        {
            path_.decreaseDimension();   
        }

        template<class WrappedScope, class NameTag, class ValueType> 
        void operator()(const ValueType&, NameTag ,EnterAttributeTag, 
                        WrappedScope)
        {
            typedef typename WrappedScope::type Scope;
            path_.prepareNext(Scope::getAttributeName(NameTag()));
        }
        template<class WrappedScope, class NameTag, class ValueType> 
        void operator()(const ValueType&, NameTag, LeaveAttributeTag, 
                        WrappedScope)
        {/* no op*/}

        template<class CollectionType, class NameTag,class WrappedScope> 
        void operator()(const CollectionType&, NameTag ,EnterCollectionTag, 
                        WrappedScope)
        {
            path_.increaseDimension();
        }
        template<class CollectionType, class NameTag,class WrappedScope> 
        void operator()(const CollectionType&, NameTag, LeaveCollectionTag, 
                        WrappedScope)
        {
            path_.decreaseDimension();
        }

        template<class WrappedScope, class NameTag, class ValueType> 
        void operator()(const ValueType& ,NameTag , EnterCollItemTag,
                        WrappedScope)
        { /* no op*/        }

        template<class WrappedScope, class NameTag, class ValueType> 
        void operator()(const ValueType& ,NameTag , LeaveCollItemTag, 
                        WrappedScope)
        {
            path_.prepareNext();
        }
    private:
        std::ostream& stream_;
        ObjectPath path_;
    };
    // If reflected attributes can be added to any class then this can only be 
    // done if the RecordType lives in a namespace where this operator can be 
    // found through Koenig lookup
    // Otherwise this template applies to anything and thus conflicts with 
    // stream operators that are defined in std.
    // 
    //template<class RecordType> std::ostream& operator<<(std::ostream& out, 
    //                                                  const RecordType& rec)
    /*{
        return print(out,rec,IsCompiletimeIterable<RecordType>::type());
    }

    template<class ValueType> std::ostream& print(std::ostream& out, 
                        const ValueType& value, NotCompileTIterableObjectTag)
    {
        return out.operator<<(out,value);
    }*/
    template<class RecordType> std::ostream& print(std::ostream& out, 
                                                const RecordType& rec)
    {
        return print(out,rec,IsCompiletimeIterable<RecordType>::type());
    }
    template<class RecordType> std::ostream& print(std::ostream& out, 
                                            const RecordType& rec, 
                                            NotCompileTIterableObjectTag)
    {
        std::ostringstream error;
        error << "Reflection not supported for: " << typeid(rec).name();
        throw std::logic_error(error.str());
        return out; 
    }
    template<class RecordType> std::ostream& print(std::ostream& out, 
                                                const RecordType& rec, 
                                                CompileTIterableObjectTag)
    {
        PrintRecord myPrinta(out);
        myPrinta.print(rec);
        return out;
    }
    template<class RecordType> std::string getRecordDesc(const RecordType& rec)
    {
        std::ostringstream recordStream;
        print(recordStream, rec);
        return recordStream.str();
    }
}

#ifdef _MSC_VER
#pragma warning(default: 4267)  
#endif

#endif // 

