////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software 
//     for any purpose. It is provided "as is" without express or implied 
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_PRIVILIGEDACCESS_H_INC_
#define REFLIB_PRIVILIGEDACCESS_H_INC_

#include <boost/type_traits/add_reference.hpp>

#include "../qualifiedtype.h"

#include "../issameclass.h"

#include "passtype.h"

namespace refl
{
    // We borrow from Smalltalk the terms ClassVariable and InstanceVariable 
    // because they are clearer then static member variable and non-static 
    // member variable
    struct InstanceVariableTag{};

    struct ClassVariableTag{};

    // traversion might be static members only 
    // (in which case no record instance is needed)
    // non-static members only or both.
    struct TraverseInstanceAndClassVariables{};

    struct KindIsMatch{};

    struct KindIsNoMatch{};

    template<class AttibuteKind, class MemberKinds> struct AttributeKindMatches
    {
        enum {memberkindsDontFilter = boost::is_same<TraverseInstanceAndClassVariables,
                                        MemberKinds>::value};  
        enum {value = boost::is_same<AttibuteKind,MemberKinds>::value};  
        typedef typename boost::ct_if<(memberkindsDontFilter != 0),
                            KindIsMatch,
                            typename boost::ct_if<(value != 0),
                                                KindIsMatch,
                                                KindIsNoMatch >::type>::type
            type;
    };
    

    namespace detail 
    {
        // check for instance Herb Sutters article on "befriending templates"
        // http://www.cuj.com/documents/s=8244/cujcexp2101sutter/sutter.htm
        // why we choose this form (Note that we do want to support VC++ 6.0)
        struct PrivilegedAccessImpl
        {
            // if no kinds are specified:
            // return references to static and non-static membervariables
            template<class ReflectedClass, class Tag, class  ValueType> static
            typename boost::add_reference<ValueType>::type
            getValueHandle(ReflectedClass& recordData, Tag tag, 
                        PassType<ValueType>)
            {
    /*          MSVC++ 6.0 chokes on that one:
                recordData.template getByTag<Tag>();
    */
                return recordData.getByTag(tag);
            }
            // return references to static OR non-static membervariables
            template<class ReflectedClass, class Tag, class  ValueType,
                    class MemberKind> static
            typename boost::add_reference<ValueType>::type
            getValueHandle(ReflectedClass& recordData, Tag tag,
                        PassType<ValueType>, MemberKind kind)
            {
                return recordData.getByTag(tag,kind);
            }
            // workaround for missing partial ordering of function
            // templates with vc6
            template<class ReflectedClass, class Tag, class  ValueType,
                    class MemberKind> static
            typename boost::add_reference<ValueType>::type
            getValueHandle(ReflectedClass& recordData, Tag tag,
                        PassType<ValueType>, MemberKind kind, IsNotSameClassTag)
            {
                return recordData.getByTag(tag,kind);
            }
            // if mixed kinds are specified:
            // return references to static and non-static membervariables
            template<class ReflectedClass, class Tag, class  ValueType> static
            typename boost::add_reference<ValueType>::type
            getValueHandle(ReflectedClass& recordData, Tag tag,
                        PassType<ValueType>,TraverseInstanceAndClassVariables, 
                        IsSameClassTag)
            {
                // if tag refers to a static member variable then getByTag is a 
                // static member function and thus cannot be overloaded on the 
                // constness of any this. 
                return recordData.getByTag(tag);
            }
            // return references to static membervariables
            template<class ReflectedClass, class Tag, class  ValueType> static
            typename boost::add_reference<ValueType>::type
            getStaticValueHandle(Tag tag, PassType<ValueType> wrappedValueType,
                             PassType<ReflectedClass> wrappedReflectedClassType)
            {
                return getStaticValueHandle(tag, wrappedValueType, 
                        wrappedReflectedClassType, IsConst<ValueType>::type());
            }
            // for static member functions the trick to overload on the 
            // constness of *this cannot work - hence the function that returns
            // a modifyable reference to a static member must be named 
            // differently (or overloaded otherwise).
            template<class ReflectedClass, class Tag, class  ValueType> static
            typename boost::add_reference<ValueType>::type
            getStaticValueHandle(Tag tag, PassType<ValueType>,
                                PassType<ReflectedClass>,TypeIsConstTag)
            {
                return ReflectedClass::getByTag(tag,ClassVariableTag());
            }
            template<class ReflectedClass, class Tag, class  ValueType> static
            typename boost::add_reference<ValueType>::type
            getStaticValueHandle(Tag tag, PassType<ValueType>,
                                PassType<ReflectedClass>,TypeIsNotConstTag)
            {
                return ReflectedClass::getRefByTag(tag,ClassVariableTag());
            }
        };
        template<class ValueType, class MemberKinds> struct PrivilegedAccess
        {
            typedef typename boost::add_reference<ValueType>::type RefType;
            template<class ReflectedClass, class Tag> static
            RefType getValueHandle(ReflectedClass& recordData, Tag tag)
            {
                // unfortunately the trick does not work with gcc 3.2.
                typedef typename IsSameClass<MemberKinds,
                                TraverseInstanceAndClassVariables>::type
                        PartialOrderingBugWorkaround;
                return PrivilegedAccessImpl::getValueHandle(recordData,
                                        tag,PassType<ValueType>(),MemberKinds(),
                                        PartialOrderingBugWorkaround());
            }
            typedef typename boost::add_reference<ValueType>::type RefType;
            template<class ReflectedClass, class Tag> static
            RefType getValueHandle(PassType<ReflectedClass> type, Tag tag, 
                                ClassVariableTag classVar)
            {
                return PrivilegedAccessImpl::getStaticValueHandle(tag,
                                PassType<ValueType>(),type);
            }
        };
    }
    template<class ReflectedClass, class MemberKinds, int Pos> struct Intruder
    {
        template<class ValueType> static 
        typename boost::add_reference<ValueType>::type 
        getValueHandle(ReflectedClass& recordData, PassType<ValueType>)
        {
            typedef typename ReflectedClass::template 
                GetNameTag<Pos,ReflectedClass::ThisClass>::type AttributeTag;
        
            typedef typename ReflectedClass::template 
                GetMemberKind<Pos,ReflectedClass::ThisClass>::type 
                AttributeKind;

            typedef detail::PrivilegedAccess<ValueType,
                                            MemberKinds> MyIntruder;
            typedef typename AttributeKindMatches<AttributeKind,
                MemberKinds>::type
                KindMatches;
            return MyIntruder::getValueHandle(recordData,AttributeTag());
        }

/*
d:\Arne\Beruf\C++\Reflection\ReflectedAttributes\vc6\priviligedaccess.h(195) : error C2440: 'return': 'const ABstatic::GetValueType<0,ABstatic::ThisClass>::const_type' kann nicht in 'boost::add_reference<T>::type' konvertiert werden
        with
        [
            T=AAstatic
        ]
        Durch die Konvertierung gehen Qualifizierer verloren
        d:\Arne\Beruf\C++\Reflection\ReflectedAttributes\vc6\priviligedaccess.h(311): Siehe Verweis auf Instanziierung der kompilierten Funktionsvorlage 'boost::add_reference<T>::type refl::Intruder<ReflectedClass,MemberKinds,Pos>::getValueHandle<AttributeValueType>(refl::PassType<T>)'
        with
        [
            T=AAstatic,
            ReflectedClass=const ConstReflectedClass,
            MemberKinds=refl::ClassVariableTag,
            Pos=0
        ]
        d:\Arne\Beruf\C++\Reflection\ReflectedAttributes\vc6\priviligedaccess.h(295): Siehe Verweis auf Instanziierung der kompilierten Funktionsvorlage 'void refl::MatchingAttributeHandler<ReflectedClass,MemberKinds,Pos,HoldsImplForMatchingKind>::processSingleAttribute<Function>(Function &,refl::KindIsMatch)'
        with
        [
            ReflectedClass=const ConstReflectedClass,
            MemberKinds=refl::ClassVariableTag,
            Pos=0,
            HoldsImplForMatchingKind=refl::DepthFirst<const ConstReflectedClass,refl::ClassVariableTag,refl::IncludeBaseClassesOfAttributesTag>::SingleAttributeHelper<0>::ThisClass,
            Function=refl::PrintRecord
        ]
        d:\Arne\Beruf\C++\Reflection\ReflectedAttributes\vc6\depthfirstforeach.h(377): Siehe Verweis auf Instanziierung der kompilierten Funktionsvorlage 'void refl::MatchingAttributeHandler<ReflectedClass,MemberKinds,Pos,HoldsImplForMatchingKind>::processSingleAttribute<Function>(Function &)'
        with
        [
            ReflectedClass=const ConstReflectedClass,
            MemberKinds=refl::ClassVariableTag,
            Pos=0,
            HoldsImplForMatchingKind=refl::DepthFirst<const ConstReflectedClass,refl::ClassVariableTag,refl::IncludeBaseClassesOfAttributesTag>::SingleAttributeHelper<0>::ThisClass,
            Function=refl::PrintRecord
        ]
        d:\Arne\Beruf\C++\Reflection\ReflectedAttributes\vc6\depthfirstforeach.h(524): Siehe Verweis auf Instanziierung der kompilierten Funktionsvorlage 'void refl::
DepthFirst<ReflectedClass,MemberKinds,IncludeBaseClassesOfAttributes>::SingleAttributeHelper<Pos>::processSingleAttribute<Function>(Function &)'
        with
        [
            ReflectedClass=const ConstReflectedClass,
            MemberKinds=refl::ClassVariableTag,
            IncludeBaseClassesOfAttributes=refl::IncludeBaseClassesOfAttributesTag,
            Pos=0,
            Function=refl::PrintRecord
        ]
        d:\Arne\Beruf\C++\Reflection\ReflectedAttributes\vc6\depthfirstforeach.h(555): Siehe Verweis auf Instanziierung der kompilierten Funktionsvorlage 'void refl::DepthFirst<ReflectedClass,MemberKinds,IncludeBaseClassesOfAttributes>::Iter<1>::process<Function>(Function &)'
        with
        [
            ReflectedClass=const ConstReflectedClass,
            MemberKinds=refl::ClassVariableTag,
            IncludeBaseClassesOfAttributes=refl::IncludeBaseClassesOfAttributesTag,
            Function=refl::PrintRecord
        ]
        d:\Arne\Beruf\C++\Reflection\ReflectedAttributes\vc6\depthfirstforeach.h(624): Siehe Verweis auf Instanziierung der kompilierten Funktionsvorlage 'void refl::DepthFirst<ReflectedClass,MemberKinds,IncludeBaseClassesOfAttributes>::doForeach<Function>(Function &)'
        with
        [
            ReflectedClass=const ConstReflectedClass,
            MemberKinds=refl::ClassVariableTag,
            IncludeBaseClassesOfAttributes=refl::IncludeBaseClassesOfAttributesTag,
            Function=refl::PrintRecord
        ]
        */
        template<class ValueType> static 
        typename boost::add_reference<ValueType>::type 
        getValueHandle(PassType<ValueType>)
        {
            typedef typename ReflectedClass::template 
                GetNameTag<Pos,ReflectedClass::ThisClass>::type AttributeTag;
        
            typedef typename ReflectedClass::template 
                GetValueType<Pos,ReflectedClass::ThisClass>::type 
                AttributeValueType;
        
            typedef typename GetQualifiedMemberType<ReflectedClass,
                        AttributeValueType>::type QualifiedValueType;

            typedef detail::PrivilegedAccess<QualifiedValueType,
                                            ClassVariableTag> MyIntruder;
            return MyIntruder::getValueHandle(PassType<ReflectedClass>(),
                                    AttributeTag(),ClassVariableTag());
        }
    };
    // having the attribute kind as a filter the need to generate no ops
    // when processing an attribute with non matching kind arises.
    // derived classes should define what happen when the kind matches:
    // template<class Function> static void
    // processMatchingAttribute(ReflectedClass& recordData,
    //                      Function& function)
    template<class ReflectedClass, class MemberKinds, int Pos, 
            class HoldsImplForMatchingKind> 
    struct MatchingAttributeHandler: Intruder<ReflectedClass,MemberKinds,Pos>
    {
        typedef Intruder<ReflectedClass,MemberKinds,Pos> Inherited;
        template<class Function> static void
        processSingleAttribute(ReflectedClass& recordData,
                                Function& function)
        {  
            typedef typename ReflectedClass::template
                GetMemberKind<Pos,ReflectedClass::ThisClass>::type 
                AttributeKind;

            typedef typename AttributeKindMatches<AttributeKind,
                                                MemberKinds>::type KindMatches;
            processSingleAttribute(recordData,function,KindMatches());
        } 
        // binary overload
        template<class Function> static void
        processSingleAttribute(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord,
                                Function& function)
        {  
            typedef typename ReflectedClass::template
                GetMemberKind<Pos,ReflectedClass::ThisClass>::type 
                AttributeKind;

            typedef typename AttributeKindMatches<AttributeKind,
                                                MemberKinds>::type KindMatches;
            processSingleAttribute(lhsRecord, rhsRecord,function,KindMatches());
        } 
        template<class Function> static void 
        processSingleAttribute(ReflectedClass&,Function&,KindIsNoMatch)
        {  
            /* no op. 
            Happens when instance variables should be traversed and 
            the current attribute is a class variable (static member variable)
            */
        }   
        // binary overload
        template<class Function> static void 
        processSingleAttribute(ReflectedClass&,ReflectedClass&,Function&,KindIsNoMatch)
        {  
            /* no op. 
            Happens when instance variables should be traversed and 
            the current attribute is a class variable (static member variable)
            */
        }   
        template<class Function> static void 
        processSingleAttribute(ReflectedClass& recordData,Function& function,
                            KindIsMatch)
        {  
            
            typedef typename ReflectedClass::template 
                GetValueType<Pos,ReflectedClass::ThisClass>::type 
                AttributeValueType;
            typedef typename GetQualifiedMemberType<ReflectedClass,
                        AttributeValueType>::type QualifiedValueType;
            HoldsImplForMatchingKind::processMatchingAttribute(recordData,
                                                              function,
                    Inherited::getValueHandle(recordData, PassType<QualifiedValueType>()));
        }   
        // binary overload
        template<class Function> static void 
        processSingleAttribute(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord,
                            Function& function,KindIsMatch)
        {  
            
            typedef typename ReflectedClass::template 
                GetValueType<Pos,ReflectedClass::ThisClass>::type 
                AttributeValueType;
            typedef typename GetQualifiedMemberType<ReflectedClass,
                        AttributeValueType>::type QualifiedValueType;
            HoldsImplForMatchingKind::processMatchingAttribute(lhsRecord,rhsRecord,
                                                              function,
                    Inherited::getValueHandle(lhsRecord, PassType<QualifiedValueType>()),
                    Inherited::getValueHandle(rhsRecord, PassType<QualifiedValueType>()));
        }   

        // This version supports access for static members only when the caller
        // has no appropriate this pointer at hand.
        template<class Function> static void 
        processSingleAttribute(Function& function)
        {  
            typedef typename ReflectedClass::template
                GetMemberKind<Pos,ReflectedClass::ThisClass>::type 
                AttributeKind;

            typedef typename AttributeKindMatches<AttributeKind,
                                                ClassVariableTag>::type 
                                                KindMatches;

            processSingleAttribute(function,KindMatches());
        } 
        template<class Function> static void 
        processSingleAttribute(Function&,KindIsNoMatch)
        {  
            /* no op. 
            Happens here when the current attribute is an instance variable 
            */
        }   
        template<class Function> static void 
        processSingleAttribute(Function& function,KindIsMatch)
        {  
            typedef typename ReflectedClass::template 
                GetValueType<Pos,ReflectedClass::ThisClass>::type 
                AttributeValueType;
            typedef typename GetQualifiedMemberType<ReflectedClass,
                        AttributeValueType>::type QualifiedValueType;
            HoldsImplForMatchingKind::processMatchingAttribute(function,
                            Inherited::getValueHandle(PassType<QualifiedValueType>()));
        }
        // Note how radically this specialization avoids using an instance
        // of ReflectedClass
    };
}


#endif //
